# -*- coding: utf-8 -*-

# Copyright (c) 2012-2021 Roel Derickx, Paul Norman <penorman@mac.com>,
# Sebastiaan Couwenberg <sebastic@xs4all.nl>, The University of Vermont
# <andrew.guertin@uvm.edu>, github contributors

# Released under the MIT license, as given in the file LICENSE, which must
# accompany any distribution of this code.

from .version import __version__

'''
ogr2osm will do the following, given an instance dw of class DataWriterBase:

    dw.open()
    try:
        dw.write_header()
        dw.write_nodes(node_list)
        dw.write_ways(way_list)
        dw.write_relations(relation_list)
        dw.write_footer()
    finally:
        dw.close()
'''

class DataWriterBase:
    def __init__(self):
        pass
    
    
    def get_version(self):
        return __version__
    
    
    def open(self):
        pass
    
    
    def write_header(self, bounds):
        pass
    
    
    def write_nodes(self, nodes):
        pass
    
    
    def write_ways(self, ways):
        pass
    
    
    def write_relations(self, relations):
        pass
    
    
    def write_footer(self):
        pass
    
    
    def close(self):
        pass




# -*- coding: utf-8 -*-

# Copyright (c) 2012-2021 Roel Derickx, Paul Norman <penorman@mac.com>,
# Sebastiaan Couwenberg <sebastic@xs4all.nl>, The University of Vermont
# <andrew.guertin@uvm.edu>, github contributors

# Released under the MIT license, as given in the file LICENSE, which must
# accompany any distribution of this code.

class TranslationBase:
    def __init__(self):
        pass
    
    
    # Override this method if you want to modify the given layer,
    # or return None if you want to suppress the layer
    def filter_layer(self, layer):
        return layer
    
    
    # Override this method if you want to modify the given feature,
    # or return None if you want to suppress the feature
    # note 1: layer_fields contains a tuple (index, field_name, field_type)
    # note 2: reproject is a function to convert the feature to 4326 projection
    # with coordinates in traditional gis order. However, do not return the
    # reprojected feature since it will be done again in ogr2osm.
    def filter_feature(self, ogrfeature, layer_fields, reproject):
        return ogrfeature
    
    
    # Override this method if you want to modify or add tags to the xml output
    def filter_tags(self, tags):
        return tags
    
    
    # This method is used to identify identical nodes for merging. By default
    # only the rounded coordinates are taken into account, but you can extend
    # this with some tags as desired. The return value should be a hashable
    # type, if you don't want to merge you can just return a counter value.
    # note: this function will not be called for nodes belonging to a way,
    # they are always identified by the tuple (rounded_x, rounded_y).
    def get_unique_node_identifier(self, rounded_x, rounded_y, tags):
        return (rounded_x, rounded_y)
    
    
    # This method is called after the creation of an OsmGeometry object. The
    # ogr feature and ogr geometry used to create the object are passed as
    # well. Note that any return values will be discarded by ogr2osm.
    def process_feature_post(self, osmgeometry, ogrfeature, ogrgeometry):
        pass
    
    
    # Override this method if you want to modify the list of nodes, ways or
    # relations, or take any additional actions right before writing the
    # objects to the OSM file. Note that any return values will be discarded
    # by ogr2osm.
    def process_output(self, osmnodes, osmways, osmrelations):
        pass




# -*- coding: utf-8 -*-

# Copyright (c) 2012-2021 Roel Derickx, Paul Norman <penorman@mac.com>,
# Sebastiaan Couwenberg <sebastic@xs4all.nl>, The University of Vermont
# <andrew.guertin@uvm.edu>, github contributors

# Released under the MIT license, as given in the file LICENSE, which must
# accompany any distribution of this code.

import logging
from osgeo import ogr
from osgeo import osr

from .osm_geometries import OsmBoundary, OsmPoint, OsmWay, OsmRelation

class OsmData:
    def __init__(self, translation, rounding_digits=7, max_points_in_way=1800, add_bounds=False):
        # options
        self.translation = translation
        self.rounding_digits = rounding_digits
        self.max_points_in_way = max_points_in_way
        self.add_bounds = add_bounds
        
        self.__bounds = OsmBoundary()
        self.__nodes = []
        self.__unique_node_index = {}
        self.__ways = []
        self.__relations = []
        self.__long_ways_from_polygons = set()


    def __get_layer_fields(self, layer):
        layer_fields = []
        layer_def = layer.GetLayerDefn()
        for i in range(layer_def.GetFieldCount()):
            field_def = layer_def.GetFieldDefn(i)
            layer_fields.append((i, field_def.GetNameRef(), field_def.GetType()))
        return layer_fields


    # This function builds up a dictionary with the source data attributes
    # and passes them to the filter_tags function, returning the result.
    def __get_feature_tags(self, ogrfeature, layer_fields, source_encoding):
        tags = {}
        for (index, field_name, field_type) in layer_fields:
            field_value = ''
            if field_type == ogr.OFTString:
                field_value = ogrfeature.GetFieldAsBinary(index).decode(source_encoding)
            else:
                field_value = ogrfeature.GetFieldAsString(index)

            tags[field_name] = field_value.strip()
        
        return self.translation.filter_tags(tags)


    def __calc_bounds(self, ogrgeometry):
        (minx, maxx, miny, maxy) = ogrgeometry.GetEnvelope()
        self.__bounds.add_envelope(minx, maxx, miny, maxy)
    
    
    def __round_number(self, n):
        return int(round(n * 10**self.rounding_digits))
    
    
    def __add_node(self, x, y, tags, is_way_member):
        rx = self.__round_number(x)
        ry = self.__round_number(y)
        
        unique_node_id = None
        if is_way_member:
            unique_node_id = (rx, ry)
        else:
            unique_node_id = self.translation.get_unique_node_identifier(rx, ry, tags)

        if unique_node_id in self.__unique_node_index:
            return self.__nodes[self.__unique_node_index[unique_node_id]]
        else:
            node = OsmPoint(x, y, tags)
            self.__unique_node_index[unique_node_id] = len(self.__nodes)
            self.__nodes.append(node)
            return node
    
    
    def __add_way(self, tags):
        way = OsmWay(tags)
        self.__ways.append(way)
        return way
    
    
    def __add_relation(self, tags):
        relation = OsmRelation(tags)
        self.__relations.append(relation)
        return relation
    
    
    def __parse_point(self, ogrgeometry, tags):
        return self.__add_node(ogrgeometry.GetX(), ogrgeometry.GetY(), tags, False)


    def __parse_linestring(self, ogrgeometry, tags):
        way = self.__add_way(tags)
        # LineString.GetPoint() returns a tuple, so we can't call parsePoint on it
        # and instead have to create the point ourself
        previous_node_id = None
        for i in range(ogrgeometry.GetPointCount()):
            (x, y, z_unused) = ogrgeometry.GetPoint(i)
            node = self.__add_node(x, y, {}, True)
            if previous_node_id == None or previous_node_id != node.id:
                way.points.append(node)
                node.addparent(way)
                previous_node_id = node.id
        return way


    def __parse_polygon(self, ogrgeometry, tags):
        # Special case polygons with only one ring. This does not (or at least
        # should not) change behavior when simplify relations is turned on.
        if ogrgeometry.GetGeometryCount() == 0:
            logging.warning("Polygon with no rings?")
        elif ogrgeometry.GetGeometryCount() == 1:
            result = self.__parse_linestring(ogrgeometry.GetGeometryRef(0), tags)
            if len(result.points) > self.max_points_in_way:
                self.__long_ways_from_polygons.add(result)
            return result
        else:
            relation = self.__add_relation(tags)
            try:
                exterior = self.__parse_linestring(ogrgeometry.GetGeometryRef(0), {})
                exterior.addparent(relation)
            except:
                logging.warning("Polygon with no exterior ring?")
                return None
            relation.members.append((exterior, "outer"))
            for i in range(1, ogrgeometry.GetGeometryCount()):
                interior = self.__parse_linestring(ogrgeometry.GetGeometryRef(i), {})
                interior.addparent(relation)
                relation.members.append((interior, "inner"))
            return relation


    def __parse_collection(self, ogrgeometry, tags):
        # OGR MultiPolygon maps easily to osm multipolygon, so special case it
        # TODO: Does anything else need special casing?
        geometry_type = ogrgeometry.GetGeometryType()
        if geometry_type in [ ogr.wkbMultiPolygon, ogr.wkbMultiPolygon25D ]:
            if ogrgeometry.GetGeometryCount() > 1:
                relation = self.__add_relation(tags)
                for polygon in range(ogrgeometry.GetGeometryCount()):
                    ext_geom = ogrgeometry.GetGeometryRef(polygon).GetGeometryRef(0)
                    exterior = self.__parse_linestring(ext_geom, {})
                    exterior.addparent(relation)
                    relation.members.append((exterior, "outer"))
                    for i in range(1, ogrgeometry.GetGeometryRef(polygon).GetGeometryCount()):
                        int_geom = ogrgeometry.GetGeometryRef(polygon).GetGeometryRef(i)
                        interior = self.__parse_linestring(int_geom, {})
                        interior.addparent(relation)
                        relation.members.append((interior, "inner"))
                return [ relation ]
            else:
               return [ self.__parse_polygon(ogrgeometry.GetGeometryRef(0), tags) ]
        elif geometry_type in [ ogr.wkbMultiLineString, ogr.wkbMultiLineString25D ]:
            geometries = []
            for linestring in range(ogrgeometry.GetGeometryCount()):
                geometries.append(self.__parse_linestring(ogrgeometry.GetGeometryRef(linestring), tags))
            return geometries
        else:
            relation = self.__add_relation(tags)
            for i in range(ogrgeometry.GetGeometryCount()):
                member = self.__parse_geometry(ogrgeometry.GetGeometryRef(i), {})
                member.addparent(relation)
                relation.members.append((member, "member"))
            return [ relation ]
    
    
    def __parse_geometry(self, ogrgeometry, tags):
        osmgeometries = []
        
        geometry_type = ogrgeometry.GetGeometryType()

        if geometry_type in [ ogr.wkbPoint, ogr.wkbPoint25D ]:
            osmgeometries.append(self.__parse_point(ogrgeometry, tags))
        elif geometry_type in [ ogr.wkbLineString, ogr.wkbLinearRing, ogr.wkbLineString25D ]:
            # ogr.wkbLinearRing25D does not exist
            osmgeometries.append(self.__parse_linestring(ogrgeometry, tags))
        elif geometry_type in [ ogr.wkbPolygon, ogr.wkbPolygon25D ]:
            osmgeometries.append(self.__parse_polygon(ogrgeometry, tags))
        elif geometry_type in [ ogr.wkbMultiPoint, ogr.wkbMultiLineString, ogr.wkbMultiPolygon, \
                                ogr.wkbGeometryCollection, ogr.wkbMultiPoint25D, \
                                ogr.wkbMultiLineString25D, ogr.wkbMultiPolygon25D, \
                                ogr.wkbGeometryCollection25D ]:
            osmgeometries.extend(self.__parse_collection(ogrgeometry, tags))
        else:
            logging.warning("Unhandled geometry, type %s" % str(geometry_type))

        return osmgeometries


    def add_feature(self, ogrfeature, layer_fields, source_encoding, reproject = lambda geometry: None):
        ogrfilteredfeature = self.translation.filter_feature(ogrfeature, layer_fields, reproject)
        if ogrfilteredfeature is None:
            return
        
        ogrgeometry = ogrfilteredfeature.GetGeometryRef()
        if ogrgeometry is None:
            return
                
        feature_tags = self.__get_feature_tags(ogrfilteredfeature, layer_fields, source_encoding)
        if feature_tags is None:
            return
        
        reproject(ogrgeometry)

        if self.add_bounds:
            self.__calc_bounds(ogrgeometry)

        osmgeometries = self.__parse_geometry(ogrgeometry, feature_tags)

        # TODO performance: run in __parse_geometry to avoid second loop
        for osmgeometry in [ geom for geom in osmgeometries if geom ]:
            self.translation.process_feature_post(osmgeometry, ogrfilteredfeature, ogrgeometry)


    def __split_way(self, way, is_way_in_relation):
        new_points = [ way.points[i:i + self.max_points_in_way] \
                               for i in range(0, len(way.points), self.max_points_in_way - 1) ]
        new_ways = [ way ] + [ OsmWay(way.get_tags()) for i in range(len(new_points) - 1) ]

        if not is_way_in_relation:
            for new_way in new_ways[1:]:
                self.__ways.append(new_way)

        for new_way, points in zip(new_ways, new_points):
            new_way.points = points
            if new_way.id != way.id:
                for point in points:
                    point.removeparent(way)
                    point.addparent(new_way)

        return new_ways


    def __merge_into_new_relation(self, way_parts):
        new_relation = self.__add_relation({})
        new_relation.members = [ (way, "outer") for way in way_parts ]
        for way in way_parts:
            way.addparent(new_relation)


    def __split_way_in_relation(self, rel, way_parts):
        way_roles = [ m[1] for m in rel.members if m[0] == way_parts[0] ]
        way_role = "" if len(way_roles) == 0 else way_roles[0]
        for way in way_parts[1:]:
            way.addparent(rel)
            rel.members.append((way, way_role))


    def split_long_ways(self):
        if self.max_points_in_way < 2:
            # pointless :-)
            return
        
        logging.debug("Splitting long ways")

        for way in self.__ways:
            is_way_in_relation = len([ p for p in way.get_parents() if type(p) == OsmRelation ]) > 0
            if len(way.points) > self.max_points_in_way:
                way_parts = self.__split_way(way, is_way_in_relation)
                if not is_way_in_relation:
                    if way in self.__long_ways_from_polygons:
                        self.__merge_into_new_relation(way_parts)
                else:
                    for rel in way.get_parents():
                        self.__split_way_in_relation(rel, way_parts)


    def process(self, datasource):
        for i in range(datasource.get_layer_count()):
            (layer, reproject) = datasource.get_layer(i)
            
            if layer:
                layer_fields = self.__get_layer_fields(layer)
                for j in range(layer.GetFeatureCount()):
                    ogrfeature = layer.GetNextFeature()
                    self.add_feature(ogrfeature, layer_fields, datasource.source_encoding, reproject)

        self.split_long_ways()


    class DataWriterContextManager:
        def __init__(self, datawriter):
            self.datawriter = datawriter
        
        def __enter__(self):
            self.datawriter.open()
            return self.datawriter
        
        def __exit__(self, exception_type, value, traceback):
            self.datawriter.close()


    def output(self, datawriter):
        self.translation.process_output(self.__nodes, self.__ways, self.__relations)
        
        with self.DataWriterContextManager(datawriter) as dw:
            dw.write_header(self.__bounds)
            dw.write_nodes(self.__nodes)
            dw.write_ways(self.__ways)
            dw.write_relations(self.__relations)
            dw.write_footer()




# -*- coding: utf-8 -*-

# Copyright (c) 2012-2021 Roel Derickx, Paul Norman <penorman@mac.com>,
# Sebastiaan Couwenberg <sebastic@xs4all.nl>, The University of Vermont
# <andrew.guertin@uvm.edu>, github contributors

# Released under the MIT license, as given in the file LICENSE, which must
# accompany any distribution of this code.

from .ogr2osm import main

main()




# -*- coding: utf-8 -*-

# Copyright (c) 2012-2021 Roel Derickx, Paul Norman <penorman@mac.com>,
# Sebastiaan Couwenberg <sebastic@xs4all.nl>, The University of Vermont
# <andrew.guertin@uvm.edu>, github contributors

# Released under the MIT license, as given in the file LICENSE, which must
# accompany any distribution of this code.

import logging, time

from .datawriter_base_class import DataWriterBase

class OsmDataWriter(DataWriterBase):
    def __init__(self, filename, never_upload=False, no_upload_false=False, never_download=False, \
                 locked=False, add_version=False, add_timestamp=False, significant_digits=9):
        self.filename = filename
        self.never_upload = never_upload
        self.no_upload_false = no_upload_false
        self.never_download = never_download
        self.locked = locked
        self.significant_digits = significant_digits
        #self.gzip_compression_level = gzip_compression_level
        self.f = None

        # Build up a dict for optional settings
        self.attributes = {}
        if add_version:
            self.attributes.update({'version':'1'})
        if add_timestamp:
            self.attributes.update({'timestamp':time.strftime('%Y-%m-%dT%H:%M:%SZ')})
    
    
    def open(self):
        #if 0 < self.gzip_compression_level < 10:
        #    import gzip
        #    self.f = gzip.open(self.filename, "wb", self.gzip_compression_level)
        #else:
        #    self.f = open(self.filename, "w", buffering = -1)
        self.f = open(self.filename, 'w', buffering = -1)


    def write_header(self, bounds):
        logging.debug("Writing file header")
        
        self.f.write('<?xml version="1.0"?>\n')
        self.f.write('<osm version="0.6" generator="ogr2osm %s"' % self.get_version())
        if self.never_upload:
            self.f.write(' upload="never"')
        elif not self.no_upload_false:
            self.f.write(' upload="false"')
        if self.never_download:
            self.f.write(' download="never"')
        if self.locked:
            self.f.write(' locked="true"')
        self.f.write('>\n')
        
        if bounds and bounds.is_valid:
            self.f.write(bounds.to_xml(self.significant_digits))
            self.f.write('\n')

    
    def __write_geometries(self, geoms):
        for osm_geom in geoms:
            self.f.write(osm_geom.to_xml(self.attributes, self.significant_digits))
            self.f.write('\n')

    
    def write_nodes(self, nodes):
        logging.debug("Writing nodes")
        self.__write_geometries(nodes)
    
    
    def write_ways(self, ways):
        logging.debug("Writing ways")
        self.__write_geometries(ways)
    
    
    def write_relations(self, relations):
        logging.debug("Writing relations")
        self.__write_geometries(relations)
    
    
    def write_footer(self):
        logging.debug("Writing file footer")
        self.f.write('</osm>')
    
    
    def close(self):
        if self.f:
            self.f.close()
            self.f = None




# -*- coding: utf-8 -*-

# Copyright (c) 2012-2021 Roel Derickx, Paul Norman <penorman@mac.com>,
# Sebastiaan Couwenberg <sebastic@xs4all.nl>, The University of Vermont
# <andrew.guertin@uvm.edu>, github contributors

# Released under the MIT license, as given in the file LICENSE, which must
# accompany any distribution of this code.

import logging, sys, time, zlib

from .osm_geometries import OsmId, OsmPoint, OsmWay, OsmRelation
from .datawriter_base_class import DataWriterBase

is_protobuf_installed = False

class PbfDataWriter(DataWriterBase):
    pass

try:
    import ogr2osm.fileformat_pb2 as fileprotobuf
    import ogr2osm.osmformat_pb2 as osmprotobuf
    
    is_protobuf_installed = True
    
    # https://wiki.openstreetmap.org/wiki/PBF_Format

    class PbfPrimitiveGroup:
        def __init__(self, add_version, add_timestamp):
            self.stringtable = {}
            self._add_string("")
            
            self._add_version = add_version
            self._version = -1
            if self._add_version:
                self._version = 1
            self._add_timestamp = add_timestamp
            self._timestamp = time.localtime(-1)
            if self._add_timestamp:
                self._timestamp = time.localtime()
            
            self.granularity = 100
            self.lat_offset = 0
            self.lon_offset = 0
            self.date_granularity = 1000
            
            self.primitive_group = osmprotobuf.PrimitiveGroup()
        
        
        # add string s to the stringtable if not yet present and returns index
        def _add_string(self, s):
            if not s in self.stringtable:
                index = len(self.stringtable)
                self.stringtable[s] = index
                return index
            else:
                return self.stringtable[s]
        
        
        # convert given latitude to value used in pbf
        def _lat_to_pbf(self, lat):
            return int((lat * 1e9 - self.lat_offset) / self.granularity)
        
        
        # convert given longitude to value used in pbf
        def _lon_to_pbf(self, lon):
            return int((lon * 1e9 - self.lon_offset) / self.granularity)
        
        
        # convert time.struct_time to value used in pbf
        def _timestamp_to_pbf(self, timestamp):
            return int(time.mktime(timestamp) * 1000 / self.date_granularity)



    class PbfPrimitiveGroupDenseNodes(PbfPrimitiveGroup):
        def __init__(self, add_version, add_timestamp):
            super(PbfPrimitiveGroupDenseNodes, self).__init__(add_version, add_timestamp)
            
            self.__last_id = 0
            self.__last_timestamp = 0
            self.__last_changeset = 0
            self.__last_lat = 0
            self.__last_lon = 0
        
        
        def add_node(self, osmpoint):
            pbftimestamp = self._timestamp_to_pbf(self._timestamp)
            pbfchangeset = 1
            pbflat = self._lat_to_pbf(osmpoint.y)
            pbflon = self._lon_to_pbf(osmpoint.x)

            self.primitive_group.dense.id.append(osmpoint.id - self.__last_id)
            
            # osmosis always requires the whole denseinfo block
            if self._add_version or self._add_timestamp:
                self.primitive_group.dense.denseinfo.version.append(self._version)
                self.primitive_group.dense.denseinfo.timestamp.append(pbftimestamp - self.__last_timestamp)
                self.primitive_group.dense.denseinfo.changeset.append(pbfchangeset - self.__last_changeset)
                self.primitive_group.dense.denseinfo.uid.append(0)
                self.primitive_group.dense.denseinfo.user_sid.append(0)
            
            self.primitive_group.dense.lat.append(pbflat - self.__last_lat)
            self.primitive_group.dense.lon.append(pbflon - self.__last_lon)
            
            self.__last_id = osmpoint.id
            self.__last_timestamp = pbftimestamp
            self.__last_changeset = pbfchangeset
            self.__last_lat = pbflat
            self.__last_lon = pbflon
            
            for (key, value) in osmpoint.get_tags().items():
                self.primitive_group.dense.keys_vals.append(self._add_string(key))
                self.primitive_group.dense.keys_vals.append(self._add_string(value))
            self.primitive_group.dense.keys_vals.append(0)



    class PbfPrimitiveGroupWays(PbfPrimitiveGroup):
        def __init__(self, add_version, add_timestamp):
            super(PbfPrimitiveGroupWays, self).__init__(add_version, add_timestamp)
        
        
        def add_way(self, osmway):
            way = osmprotobuf.Way()
            way.id = osmway.id
            
            for (key, value) in osmway.get_tags().items():
                way.keys.append(self._add_string(key))
                way.vals.append(self._add_string(value))
            
            # osmosis always requires the whole info block
            if self._add_version or self._add_timestamp:
                way.info.version = self._version
                way.info.timestamp = self._timestamp_to_pbf(self._timestamp)
                way.info.changeset = 1
                way.info.uid = 0
                way.info.user_sid = 0
            
            prev_node_id = 0
            for node in osmway.points:
                way.refs.append(node.id - prev_node_id)
                prev_node_id = node.id
            
            self.primitive_group.ways.append(way)
        
        

    class PbfPrimitiveGroupRelations(PbfPrimitiveGroup):
        def __init__(self, add_version, add_timestamp):
            super(PbfPrimitiveGroupRelations, self).__init__(add_version, add_timestamp)
        
        
        def add_relation(self, osmrelation):
            relation = osmprotobuf.Relation()
            relation.id = osmrelation.id
            
            for (key, value) in osmrelation.get_tags().items():
                relation.keys.append(self._add_string(key))
                relation.vals.append(self._add_string(value))
            
            # osmosis always requires the whole info block
            if self._add_version or self._add_timestamp:
                relation.info.version = self._version
                relation.info.timestamp = self._timestamp_to_pbf(self._timestamp)
                relation.info.changeset = 1
                relation.info.uid = 0
                relation.info.user_sid = 0
            
            prev_member_id = 0
            for member in osmrelation.members:
                relation.memids.append(member.id - prev_member_id)
                prev_member_id = member.id
                
                relation_type = osmprotobuf.MemberType.NODE
                if type(member) == OsmWay:
                    relation_type = osmprotobuf.MemberType.WAY
                elif type(member) == OsmRelation:
                    relation_type = osmprotobuf.MemberType.RELATION
                relation.types.append(relation_type)
            
            self.primitive_group.relations.append(relation)



    class PbfDataWriter(DataWriterBase):
        def __init__(self, filename, add_version=False, add_timestamp=False):
            self.filename = filename
            self.add_version = add_version
            self.add_timestamp = add_timestamp
            
            self.__max_nodes_per_node_block = 8000
            self.__max_node_refs_per_way_block = 32000
            self.__max_member_refs_per_relation_block = 32000
        
        
        def open(self):
            self.f = open(self.filename, 'wb', buffering = -1)
        
        
        def __write_blob(self, data, block_type):
            #logging.debug("Writing blob, type = %s" % block_type)
            
            blob = fileprotobuf.Blob()
            blob.raw_size = len(data)
            blob.zlib_data = zlib.compress(data)

            blobheader = fileprotobuf.BlobHeader()
            blobheader.type = block_type
            blobheader.datasize = blob.ByteSize()
            
            blobheaderlen = blobheader.ByteSize().to_bytes(4, byteorder='big')
            self.f.write(blobheaderlen)
            self.f.write(blobheader.SerializeToString())
            self.f.write(blob.SerializeToString())


        def write_header(self, bounds):
            logging.debug("Writing file header")
            
            header_block = osmprotobuf.HeaderBlock()
            if bounds and bounds.is_valid:
                header_block.bbox.left = int(bounds.minlon * 1e9)
                header_block.bbox.right = int(bounds.maxlon * 1e9)
                header_block.bbox.top = int(bounds.maxlat * 1e9)
                header_block.bbox.bottom = int(bounds.minlat * 1e9)
            header_block.required_features.append("OsmSchema-V0.6")
            header_block.required_features.append("DenseNodes")
            header_block.writingprogram = "ogr2osm %s" % self.get_version()
            self.__write_blob(header_block.SerializeToString(), "OSMHeader")


        def __write_primitive_block(self, pbf_primitive_group):
            #logging.debug("Primitive block generation")
            
            primitive_block = osmprotobuf.PrimitiveBlock()
            # add stringtable
            for (string, index) in sorted(pbf_primitive_group.stringtable.items(), key=lambda kv: kv[1]):
                primitive_block.stringtable.s.append(string.encode('utf-8'))
            # add geometries
            primitive_block.primitivegroup.append(pbf_primitive_group.primitive_group)
            # set parameters
            primitive_block.granularity = pbf_primitive_group.granularity
            primitive_block.lat_offset = pbf_primitive_group.lat_offset
            primitive_block.lon_offset = pbf_primitive_group.lon_offset
            primitive_block.date_granularity = pbf_primitive_group.date_granularity
            # write primitive block
            self.__write_blob(primitive_block.SerializeToString(), "OSMData")

        
        def write_nodes(self, nodes):
            logging.debug("Writing nodes")
            for i in range(0, len(nodes), self.__max_nodes_per_node_block):
                primitive_group = PbfPrimitiveGroupDenseNodes(self.add_version, self.add_timestamp)
                for node in nodes[i:i+self.__max_nodes_per_node_block]:
                    primitive_group.add_node(node)
                self.__write_primitive_block(primitive_group)
        
        
        def write_ways(self, ways):
            logging.debug("Writing ways")
            amount_node_refs = 0
            primitive_group = None
            for way in ways:
                if amount_node_refs == 0:
                    primitive_group = PbfPrimitiveGroupWays(self.add_version, self.add_timestamp)
                primitive_group.add_way(way)
                amount_node_refs += len(way.points)
                if amount_node_refs > self.__max_node_refs_per_way_block:
                    self.__write_primitive_block(primitive_group)
                    amount_node_refs = 0
            else:
                if amount_node_refs > 0:
                    self.__write_primitive_block(primitive_group)
        
        
        def write_relations(self, relations):
            logging.debug("Writing relations")
            amount_member_refs = 0
            primitive_group = None
            for relation in relations:
                if amount_member_refs == 0:
                    primitive_group = PbfPrimitiveGroupRelations(self.add_version, self.add_timestamp)
                primitive_group.add_relation(relation)
                amount_member_refs += len(relation.members)
                if amount_member_refs > self.__max_member_refs_per_relation_block:
                    self.__write_primitive_block(primitive_group)
                    amount_member_refs = 0
            else:
                if amount_member_refs > 0:
                    self.__write_primitive_block(primitive_group)
        
        
        def close(self):
            if self.f:
                self.f.close()
                self.f = None
except:
    pass




# -*- coding: utf-8 -*-

# Copyright (c) 2012-2021 Roel Derickx, Paul Norman <penorman@mac.com>,
# Sebastiaan Couwenberg <sebastic@xs4all.nl>, The University of Vermont
# <andrew.guertin@uvm.edu>, github contributors

# Released under the MIT license, as given in the file LICENSE, which must
# accompany any distribution of this code.

import logging
from lxml import etree

class OsmId:
    element_id_counter = 0
    element_id_counter_incr = -1

    @staticmethod
    def set_id(start_id, is_positive = False):
        OsmId.element_id_counter = start_id
        if is_positive:
            OsmId.element_id_counter_incr = 1


    @staticmethod
    def load_id(filename):
        with open(filename, 'r') as ff:
            OsmId.element_id_counter = int(ff.readline(20))
        logging.info("Starting counter value '%d' read from file '%s'." % \
                     (OsmId.element_id_counter, filename))


    @staticmethod
    def save_id(filename):
        with open(filename, 'w') as ff:
            ff.write(str(OsmId.element_id_counter))
        logging.info("Wrote elementIdCounter '%d' to file '%s'" % \
                     (OsmId.element_id_counter, filename))



class OsmBoundary:
    def __init__(self):
        self.is_valid = False
        self.minlon = 0.0
        self.maxlon = 0.0
        self.minlat = 0.0
        self.maxlat = 0.0


    def add_envelope(self, minx, maxx, miny, maxy):
        if self.is_valid:
            self.minlon = min(self.minlon, minx)
            self.maxlon = max(self.maxlon, maxx)
            self.minlat = min(self.minlat, miny)
            self.maxlat = max(self.maxlat, maxy)
        else:
            self.is_valid = True
            self.minlon = minx
            self.maxlon = maxx
            self.minlat = miny
            self.maxlat = maxy


    def to_xml(self, significant_digits):
        xmlattrs = { 'minlon':(('%%.%df' % significant_digits) % self.minlon).strip('0'), \
                     'minlat':(('%%.%df' % significant_digits) % self.minlat).strip('0'), \
                     'maxlon':(('%%.%df' % significant_digits) % self.maxlon).strip('0'), \
                     'maxlat':(('%%.%df' % significant_digits) % self.maxlat).strip('0') }
        xmlobject = etree.Element('bounds', xmlattrs)
        
        return etree.tostring(xmlobject, encoding='unicode')



class OsmGeometry:
    def __init__(self):
        self.id = self.__get_new_id()
        self.__tags = {}
        self.__parents = set()
    
    
    def __get_new_id(self):
        OsmId.element_id_counter += OsmId.element_id_counter_incr
        return OsmId.element_id_counter
    
    
    def add_tags(self, tags):
        self.__tags.update(tags)
    
    
    def get_tags(self):
        return self.__tags
    
    
    def addparent(self, parent):
        self.__parents.add(parent)


    def removeparent(self, parent):
        self.__parents.discard(parent)
    
    
    def get_parents(self):
        return self.__parents
    
    
    def to_xml(self, attributes, significant_digits):
        pass



class OsmPoint(OsmGeometry):
    def __init__(self, x, y, tags):
        super().__init__()
        self.x = x
        self.y = y
        self.add_tags(tags)
    
    
    def to_xml(self, attributes, significant_digits):
        xmlattrs = { 'visible':'true', \
                     'id':('%d' % self.id), \
                     'lat':(('%%.%df' % significant_digits) % self.y).strip('0'), \
                     'lon':(('%%.%df' % significant_digits) % self.x).strip('0') }
        xmlattrs.update(attributes)

        xmlobject = etree.Element('node', xmlattrs)

        for (key, value) in self.get_tags().items():
            tag = etree.Element('tag', { 'k':key, 'v':value })
            xmlobject.append(tag)
        
        return etree.tostring(xmlobject, encoding='unicode')



class OsmWay(OsmGeometry):
    def __init__(self, tags):
        super().__init__()
        self.points = []
        self.add_tags(tags)


    def to_xml(self, attributes, significant_digits):
        xmlattrs = { 'visible':'true', 'id':('%d' % self.id) }
        xmlattrs.update(attributes)

        xmlobject = etree.Element('way', xmlattrs)

        for node in self.points:
            nd = etree.Element('nd', { 'ref':('%d' % node.id) })
            xmlobject.append(nd)
        for (key, value) in self.get_tags().items():
            tag = etree.Element('tag', { 'k':key, 'v':value })
            xmlobject.append(tag)

        return etree.tostring(xmlobject, encoding='unicode')



class OsmRelation(OsmGeometry):
    def __init__(self, tags):
        super().__init__()
        self.members = []
        self.add_tags(tags)


    def to_xml(self, attributes, significant_digits):
        xmlattrs = { 'visible':'true', 'id':('%d' % self.id) }
        xmlattrs.update(attributes)

        xmlobject = etree.Element('relation', xmlattrs)

        for (member, role) in self.members:
            member = etree.Element('member', { 'type':'way', 'ref':('%d' % member.id), 'role':role })
            xmlobject.append(member)

        tag = etree.Element('tag', { 'k':'type', 'v':'multipolygon' })
        xmlobject.append(tag)
        for (key, value) in self.get_tags().items():
            tag = etree.Element('tag', { 'k':key, 'v':value })
            xmlobject.append(tag)

        return etree.tostring(xmlobject, encoding='unicode')

